open Unix
open Pervasives

open Types
open Utils

exception Auth_failed

(* port on which the server is listening *)
let port = ref 8080

(* from where can we connect *)
let connect_from = ref None

(* Is there a following proxy *)
let proxy = ref None
let noProxy = ref []

(* what are the headers we don't want *)
let forbidden_headers = ref ["proxy*"]

(* list of the different profiles *)
let prof_list = ref []

(* give the profile for this authentication *)
let get_profile auth =
  let rec aux = function
    | [] -> raise Auth_failed
    | a::l ->
        if a.auth_pass = auth
        then a
        else aux l in
  aux !prof_list

(* fill a list *)
let read_list list name conf =
  let tok = ref (String.lowercase (next_token conf)) in
  while !tok <> "" && !tok <> "end" do
    list := !tok :: !list;
    tok := String.lowercase (next_token conf)
  done;
  if !tok = "" then
    (output_string stderr
       ("Invalid configuration file (" ^ name ^ " without end)\n");
     exit 1)

(* read the configuration file *)
let init conf_file =
  (* read the conf file *)
  let conf_file =
    if conf_file <> ""
    then conf_file
    else
      if Sys.file_exists "swiftsurf.conf"
      then "swiftsurf.conf"
      else
        (let home_file = Filename.concat (Sys.getenv "HOME") ".swiftsurf" in
        if Sys.file_exists home_file
        then home_file
        else
          if Sys.file_exists "/etc/swiftsurf.conf"
          then "/etc/swiftsurf.conf"
          else "/dev/null") in
  (* set up profiles *)
  let default_prof = {
    auth_pass = "";
    forbidden_urls = [];
    allowed_urls = [];
    canGET = true;
    canPOST = true;
    canCONNECT = true;
    allCONNECT = false;
    req_1 = false;
    req_2 = false;
    req_in = false;
    req_out = false;
    ans_1 = false;
    ans_2 = false;
    ans_in = false;
    ans_out = false } in
  let current_prof = ref default_prof in
  let conf = open_in conf_file in
  try
    while true do
      match next_token conf with
      | "logFile" ->
          let fds = Unix.openfile (next_token conf)
              [Unix.O_WRONLY; Unix.O_TRUNC; O_CREAT] 420 in
          Unix.dup2 fds Unix.stdout;
          Unix.close fds
      | "port" ->
          port := int_of_string (next_token conf)
      | "connectFrom" ->
          connect_from := Some (Str.regexp (next_token conf))
      | "profile" ->
          current_prof := {
            auth_pass = next_token conf;
            forbidden_urls = [];
            allowed_urls = [];
            canGET = true;
            canPOST = true;
            canCONNECT = true;
            allCONNECT = false;
            req_1 = false;
            req_2 = false;
            req_in = false;
            req_out = false;
            ans_1 = false;
            ans_2 = false;
            ans_in = false;
            ans_out = false }
      | "endProfile" ->
          if !current_prof == default_prof
          then
            (output_string stderr
               ("Invalid configuration file (endProfile without profile)\n");
             exit 1)
          else
            (prof_list := !current_prof :: !prof_list;
             current_prof := default_prof)
      | "timeout" ->
          Types.timeout := float_of_string (next_token conf);
          if !Types.timeout=0. then Types.timeout := 99999.;
      | "stats" ->
          Types.stats := int_of_string (next_token conf)
      | "debug" ->
          Types.debug := int_of_string (next_token conf)
      | "noGET" ->
          !current_prof.canGET <- false
      | "noPOST" ->
          !current_prof.canPOST <- false
      | "noCONNECT" ->
          !current_prof.canCONNECT <- false
      | "allCONNECT" ->
          !current_prof.allCONNECT <- false
      | "request" ->
          let req = int_of_string (next_token conf) in
          !current_prof.req_1 <- ((req land 1) = 1);
          !current_prof.req_2 <- ((req land 2) = 2);
          !current_prof.req_in <- ((req land 4) = 4);
          !current_prof.req_out <- ((req land 8) = 8)
      | "answer" ->
          let ans = int_of_string (next_token conf) in
          !current_prof.ans_1 <- ((ans land 1) = 1);
          !current_prof.ans_2 <- ((ans land 2) = 2);
          !current_prof.ans_in <- ((ans land 4) = 4);
          !current_prof.ans_out <- ((ans land 8) = 8)
      | "maxConns" ->
          Types.max_conns := int_of_string (next_token conf)
      | "nextProxy" ->
          let p_host = next_token conf in
          let p_port = int_of_string (next_token conf) in
          let entry = gethostbyname p_host in
          proxy := Some (ADDR_INET (entry.h_addr_list.(0), p_port))
      | "noProxy" ->
          read_list noProxy "noProxy" conf
      | "forbiddenUrl" ->
          let l = ref [] in
          read_list l "forbiddenUrl" conf;
          !current_prof.forbidden_urls <- !l
      | "forbiddenFile" ->
          let forb = open_in (next_token conf) in
          let tok = ref (String.lowercase (next_token forb)) in
          (try
            while true do
              !current_prof.forbidden_urls <-
                !tok :: !current_prof.forbidden_urls;
              tok := String.lowercase (next_token forb)
            done
          with _ ->
            close_in forb)
      | "allowedUrl" ->
          let l = ref [] in
          read_list l "allowedUrl" conf;
          !current_prof.allowed_urls <- !l
      | "forbiddenHeader" ->
          read_list forbidden_headers "forbiddenHeader" conf
      | "answerGIF" ->
          let image = next_token conf in
          Types.forbid_url := "HTTP/1.0 200 OK\r\nServer: Proxy http filtrant\r\nAccept-Ranges: bytes\r\nConnection: close\r\nContent-Type: image/gif\r\n\r\n";
          Types.forbid_url := !Types.forbid_url ^ (read_file image)
      | s ->
          output_string stderr
            ("Invalid configuration file (unknown token " ^ s ^ ")\n");
          exit 1
    done
  with _ ->
    close_in conf;
    if !prof_list == []
    then prof_list := [default_prof]

(* Does this site need a proxy *)
let need_proxy site =
  let rec need_aux = function
    | [] ->
        !proxy
    | pat::l ->
        if match_pattern site pat
        then None
        else need_aux l
  in
  need_aux !noProxy

(* see if a string matches a pattern in the list *)
let rec match_list str = function
  | [] -> false
  | pat::l ->
      match_pattern str pat || match_list str l

(* test if an url is allowed or not *)
let ok_url url prof =
  if prof.allowed_urls = []
  then
    not (match_list url prof.forbidden_urls)
  else
    (match_list url prof.allowed_urls)
      && not (match_list url prof.forbidden_urls)

(* test if this line of header is allowed *)
let ok_header header =
  not (match_list header !forbidden_headers)
