open Types
open Activebuffer

(****************************)
(* Operations on connexions *)
(****************************)

(* bind the server socket *)
let bind_server port =
  let fds = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  Unix.setsockopt fds Unix.SO_REUSEADDR true;
  Unix.bind fds (Unix.ADDR_INET (Unix.inet_addr_any, port));
  Unix.listen fds 5;
  fds

(* catch all exceptions made by a function *)
let trywith f =
  try
    f ()
  with _ -> ()

(* Closes a connexion *)
let close_connexion conn =
  (match conn.state with
  | DNS -> Dns.forget conn
  | _ -> ());
  trywith
    (fun () -> Unix.shutdown conn.client Unix.SHUTDOWN_ALL);
  trywith
    (fun () -> Unix.close conn.client);
  if conn.server <> Unix.stdin then
    trywith (fun () -> Unix.close conn.server)

(* put a connexion in a state where it just has to
 * finish writing data to the client and then exit *)
let finish conn str =
  (match conn.state with
  | DNS -> Dns.forget conn
  | _ -> ());
  if str <> ""
  then conn.write_ans <- Activebuffer.activebuffer_of_string str
  else
    if conn.state_ans == CMD_LINE
    then conn.write_ans <- Activebuffer.activebuffer_of_string not_avail;
  if conn.server <> Unix.stdin then
    (trywith (fun () -> Unix.close conn.server);
     conn.server <- Unix.stdin);
  conn.state <- FINISHING

(***********************)
(* Operations on files *)
(***********************)

(* read a whole file *)
let read_file name =
  let chan = open_in_bin name in
  let buf_size = ref 1024 in
  let buf = ref (String.create !buf_size) in
  let nb_read = ref 0 in
  let cont = ref true in
  while !cont do
    let tmp = input chan !buf !nb_read (!buf_size - !nb_read) in
    if tmp = 0
    then cont := false
    else
      (nb_read := !nb_read + tmp;
       if !nb_read = !buf_size
       then
         (buf := !buf ^ (String.create !buf_size);
          buf_size := 2 * !buf_size))
  done;
  close_in chan;
  String.sub !buf 0 !nb_read

(* Give the next token of a channel
 * Commentaries go from # to the end of the line  
 * this is used for reading the configuration file
 *)
let next_token chan =
  let res = ref "" in
  let ch = ref (input_char chan) in
  while !ch = ' ' || !ch = '\n' || !ch = '\r'
        || !ch = '\t' || !ch = '#' do
    if !ch = '#'
    then (let _ = input_line chan in ());
    ch := input_char chan
  done;
  while not (!ch = ' ' || !ch = '\n' || !ch = '\r'
        || !ch = '\t' || !ch = '#') do
    res := !res ^ (String.make 1 !ch);
    ch := input_char chan
  done;
  !res

(* does the string match the pattern *)
let match_pattern str pat =
  let size_s = String.length str in
  let size_p = String.length pat in
  let rec match_aux pos_s pos_p =
    if pos_p == size_p then
      pos_s == size_s
    else
      ((pat.[pos_p] == '*' && match_aux pos_s (pos_p+1))
     ||
       (pos_s != size_s &&
        (match pat.[pos_p] with
        | '*' ->
            match_aux (pos_s+1) pos_p
        | '?' -> match_aux (pos_s+1) (pos_p+1)
        | x -> x = (Char.lowercase str.[pos_s])
              && match_aux (pos_s+1) (pos_p+1))))
  in
  match_aux 0 0

(* decode password for authentication *)
let decode64 s =
  let val64 c =
    match c with
    | 'A' .. 'Z' -> (Char.code c) - (Char.code 'A')
    | 'a' .. 'z' -> (Char.code c) - (Char.code 'a') + 26
    | '0' .. '9' -> (Char.code c) - (Char.code '0') + 52
    | '+' -> 62 | '/' -> 63 | '=' -> 0
    | _ -> failwith "not a base64 string" in
  let len = String.length s in
  let len_res = len * 3 / 4 in
  let res = String.create len_res in
  for i=0 to len/4 - 1 do
    let i1 = 4*i and i2 = 3*i in
    let v1 = (val64 s.[i1]) lsl 18 in
    let v2 = (val64 s.[i1 + 1]) lsl 12 in
    let v3 = (val64 s.[i1 + 2]) lsl 6 in
    let v4 = val64 s.[i1 + 3] in
    let v = v1 lor v2 lor v3 lor v4 in
    res.[i2] <- Char.chr (v lsr 16);
    res.[i2 + 1] <- Char.chr (v lsr 8 land 0xFF);
    res.[i2 + 2] <- Char.chr (v land 0xFF)
  done;
  let nb_keep =
    if s.[len-1] = '=' then
      if s.[len-2] = '=' then len_res - 2 else len_res - 1
    else len_res in
  String.sub res 0 nb_keep
